﻿using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Diagnostics;
using EmailClient.Data;
using EmailClient.DomainModel.Component;

namespace EmailClient.DomainModel
{
    public interface IAccountModel
    {
        ObservableCollection<Folder> Folders { get; }

        Folder? DraftFolder { get; }

        void RefreshFolder(Folder f);

        void RefreshAll();

        Email? CreateDraft(string? subject = null, string? content = null, string? recipient = null);

        Email? SendEmail(Email draft, bool deleteDraft = false);

        void DeleteDraft(Email draft);
    }

    public class Account : ObservableObject, IAccountModel
    {
        private readonly MailServer _server;
        private bool _updateLock = false;

        private DateTimeOffset _lastSync;

        private readonly string _draftsFolderName = "Drafts";
        private readonly string _sentFolderName = "Sent";
        private int _draftsFolderId = -1;
        private int _sentFolderId = -1;

        public Account(MailServer server)
        {
            _server = server;
            Folders = new ObservableCollection<Folder>();
            (var folders, _lastSync) = _server.GetFolders();
            foreach (var folderId in folders!)
            {
                var f = CreateFolderInstance(folderId);
                if (f != null)
                    Folders.Add(f);
            }

            Folders.CollectionChanged += SubFoldersChanged;
        }

        public string Address => _server.Address;

        public ObservableCollection<Folder> Folders { get; }

        public Folder? DraftFolder => FindFolder(_draftsFolderId);

        public void RefreshFolder(Folder f)
        {
            var updateOld = _updateLock;
            _updateLock = true;
            try
            {
                if (_server.HasFolderChanged(f.Id, f.LastSync))
                    if (!UpdateFolderInstance(f))
                    {
                        RemoveFolder(f);
                        return;
                    }

                RefreshFolderEmails(f.Emails);
            }
            finally
            {
                _updateLock = updateOld;
            }
        }

        private void RefreshFolderEmails(ObservableCollection<Email> emails)
        {
            for (var i = 0; i < emails.Count; i++)
            {
                var e = emails[i];
                if (_server.HasEmailChanged(e.Id, e.LastSync))
                    if (!UpdateEmailInstance(e))
                    {
                        emails.RemoveAt(i);
                        --i;
                        e.PropertyChanged -= EmailChanged;
                    }
            }
        }

        public void RefreshAll()
        {
            var updateOld = _updateLock;
            _updateLock = true;
            try
            {
                if (_server.HasFolderChanged(-1, _lastSync))
                {
                    var (folders, newLast) = _server.GetFolders();
                    if (newLast == DateTimeOffset.MinValue)
                        throw new Exception("Account deleted!");
                    UpdateSubFolders(Folders, folders!.ToList());
                    _lastSync = newLast;
                }
                var toRefresh = new List<(ObservableCollection<Folder> Folders, int Index)> { (Folders, 0) };
                while (toRefresh.Count > 0)
                {
                    var (folders, index) = toRefresh.Last();
                    var folder = folders[index];
                    if (folder.SubFolders.Count > 0)
                    {
                        toRefresh.Add((folder.SubFolders, 0));
                        continue;
                    }
                    do
                    {

                        if (_server.HasFolderChanged(folder.Id, folder.LastSync))
                        {
                            if (!UpdateFolderInstance(folder))
                            {
                                folders.RemoveAt(index);
                                --index;
                                RemoveFolderChangeHandlers(folder);
                            }
                            else
                                RefreshFolderEmails(folder.Emails);

                        }

                        ++index;
                        if (index < folders.Count)
                        {
                            toRefresh[^1] = toRefresh[^1] with { Index = index };
                            break;
                        }

                        toRefresh.RemoveAt(toRefresh.Count - 1);
                        if (toRefresh.Count == 0)
                            break;
                        (folders, index) = toRefresh.Last();
                        folder = folders[index];
                    } while (true);
                }
            }
            finally
            {
                _updateLock = updateOld;
            }
        }

        public Email? CreateDraft(string? subject = null, string? content = null, string? recipient = null)
        {
            var draftFolder = FindFolder(_draftsFolderId);
            if (draftFolder == null) return null;
            bool updateOld = _updateLock;
            _updateLock = true;
            try
            {
                var ((emailId, folderLast), folderNew) = _server.AddEmail(_draftsFolderId);
                
                _server.SetEmailSender(emailId, _server.Address);
                if (!string.IsNullOrEmpty(recipient))
                    _server.SetEmailRecipient(emailId, recipient);
                if (!string.IsNullOrEmpty(subject))
                    _server.SetEmailSubject(emailId, subject);
                if (!string.IsNullOrEmpty(content))
                    _server.SetEmailContent(emailId, content);
                var email = CreateEmailInstance(emailId);
                if (email != null)
                    draftFolder.Emails.Add(email);
                if (!CheckFolderSync(draftFolder, folderLast, folderNew))
                    return null;
                return email;
            }
            finally
            {
                _updateLock = updateOld;
            }
        }

        public Email? SendEmail(Email draft, bool deleteDraft = false)
        {
            var updateOld = _updateLock;
            _updateLock = true;
            try
            {
                 var email = SendEmailImpl(draft);
                 if (email == null) return null;
                 if (deleteDraft)
                     DeleteDraft(draft);
                 return email;
            }
            finally
            {
                _updateLock = updateOld;
            }
        }

        public void DeleteDraft(Email draft)
        {
            var updateOld = _updateLock;
            _updateLock = true;

            try
            {
                var draftFolder = FindFolder(_draftsFolderId);
                var (folderLast, folderNew) = _server.DeleteEmail(draft.Id);
                if (draftFolder == null) return;
                draftFolder.Emails.Remove(draft);
                CheckFolderSync(draftFolder, folderLast, folderNew);
            }
            finally
            {
                _updateLock = updateOld;
            }
        }

        private Email? SendEmailImpl(Email draft)
        {
            var sentFolder = FindFolder(_sentFolderId);
            if (sentFolder == null)
            {
                RefreshAll();
                if ((sentFolder = FindFolder(_sentFolderId)) == null) return null;
            }

            var ((emailId, folderLast), folderNew) =
                _server.SendEmail(draft.Subject, draft.Content, draft.Recipient);
            var email = CreateEmailInstance(emailId);
            if (email != null)
                sentFolder.Emails.Add(email);
            if (!CheckFolderSync(sentFolder, folderLast, folderNew))
                return null;
            return email;
        }

        private Folder? CreateFolderInstance(int folderId)
        {
            do
            {
                var (name, last) = _server.GetFolderName(folderId);
                if (last == DateTimeOffset.MinValue) return null;
                if (name == _draftsFolderName) 
                    _draftsFolderId = folderId;
                else if (name == _sentFolderName)
                    _sentFolderId = folderId;
                var (subFolders, newLast) = _server.GetSubFolders(folderId);
                if (newLast == DateTimeOffset.MinValue) return null;
                if (last != newLast) continue;
                (var emails, newLast) = _server.GetFolderEmails(folderId);
                if (newLast == DateTimeOffset.MinValue) return null;
                if (last != newLast) continue;
                var f = new Folder(folderId, name!, last,
                    emails!.Select(CreateEmailInstance).Where(e => e != null).Select(e => e!),
                    subFolders!.Select(CreateFolderInstance).Where(sf => sf != null).Select(sf => sf!));
                AddFolderChangeHandlers(f);
                return f;
            } while (true);
        }

        private void AddFolderChangeHandlers(Folder f)
        {
            f.PropertyChanged += FolderChanged;
            f.SubFolders.CollectionChanged += SubFoldersChanged;
            f.Emails.CollectionChanged += EmailsChanged;
        }

        private void RemoveFolderChangeHandlers(Folder f)
        {
            f.PropertyChanged -= FolderChanged;
            f.SubFolders.CollectionChanged -= SubFoldersChanged;
            f.Emails.CollectionChanged -= EmailsChanged;
        }

        private bool UpdateFolderInstance(Folder f)
        {
            do
            {
                (var name, f.LastSync) = _server.GetFolderName(f.Id);
                if (f.LastSync == DateTimeOffset.MinValue) return false;
                f.Name = name!;
                var (emails, newLast) = _server.GetFolderEmails(f.Id);
                if (newLast == DateTimeOffset.MinValue) return false;
                if (f.LastSync != newLast) continue;
                UpdateFolderEmails(f.Emails, emails!.ToList());
                (var subFolders, newLast) = _server.GetSubFolders(f.Id);
                if (newLast == DateTimeOffset.MinValue) return false;
                if (f.LastSync != newLast) continue;
                UpdateSubFolders(f.SubFolders, subFolders!.ToList());
                return true;
            } while (true);
        }

        private void UpdateSubFolders(ObservableCollection<Folder> subFolders, IList<int> sfIds)
        {
            //Remove folders
            for (int i = 0; i < subFolders.Count; i++)
            {
                if (!sfIds.Contains(subFolders[i].Id))
                {
                    subFolders.RemoveAt(i);
                    ClearFolder(subFolders[i]);
                    --i;
                }
            }
            //Add new folders
            foreach (var sfId in sfIds.Where(fid => subFolders.All(f => f.Id != fid)))
            {
                var subF = CreateFolderInstance(sfId);
                if (subF != null) subFolders.Add(subF);
            }
        }

        private void ClearFolder(Folder f)
        {
            List<Folder> toClean = new List<Folder>() { f };
            while (toClean.Count > 0)
            {
                Folder sf = toClean[^1];
                toClean.RemoveAt(toClean.Count - 1);
                RemoveFolderChangeHandlers(sf);
                foreach (var e in sf.Emails)
                    e.PropertyChanged -= EmailChanged;
                sf.Emails.Clear();
                toClean.AddRange(sf.SubFolders.Reverse());
                sf.SubFolders.Clear();
            }
        }

        private void RemoveFolder(Folder folder)
        {
            ObservableCollection<Folder>? folders = FindFolderCollection(folder);
            folders?.Remove(folder);
            ClearFolder(folder);
        }

        private ObservableCollection<Folder>? FindFolderCollection(Folder folder)
        {
            Stack<ObservableCollection<Folder>> toSearch = new Stack<ObservableCollection<Folder>>();
            toSearch.Push(Folders);
            while (toSearch.Count > 0)
            {
                var current = toSearch.Pop();
                if (current.Contains(folder))
                    return current;
                foreach (var f in current.Where(fo => fo.SubFolders.Count > 0))
                    toSearch.Push(f.SubFolders);
            }

            return null;
        }

        private void UpdateFolderEmails(ObservableCollection<Email> emails, IList<int> eIds)
        {
            //Remove emails
            for (int i = 0; i < emails.Count; --i)
            {
                if (!eIds.Contains(emails[i].Id))
                {
                    emails[i].PropertyChanged -= EmailChanged;
                    emails.RemoveAt(i);
                    --i;
                }
            }
            //Add new emails
            foreach (var emailId in eIds.Where(eid => emails.All(e => e.Id != eid)))
            {
                var email = CreateEmailInstance(emailId);
                if (email != null)
                    emails.Add(email);
            }
        }

        private Email? CreateEmailInstance(int emailId)
        {
            do
            {
                var (sender, last) = _server.GetEmailSender(emailId);
                if (last == DateTimeOffset.MinValue) return null;
                var (recipient, newLast) = _server.GetEmailRecipient(emailId);
                if (newLast == DateTimeOffset.MinValue) return null;
                if (last != newLast) continue;
                (var subject, newLast) = _server.GetEmailSubject(emailId);
                if (newLast == DateTimeOffset.MinValue) return null;
                if (last != newLast) continue;
                (var content, newLast) = _server.GetEmailContent(emailId);
                if (newLast == DateTimeOffset.MinValue) return null;
                if (last != newLast) continue;
                (var timestamp, newLast) = _server.GetEmailTimestamp(emailId);
                if (newLast == DateTimeOffset.MinValue) return null;
                if (last != newLast) continue;
                (var status, newLast) = _server.GetEmailStatus(emailId);
                if (newLast == DateTimeOffset.MinValue) return null;
                if (last != newLast) continue;
                var e = new Email(emailId, sender!, recipient!, subject!, content!, timestamp, status, last);
                e.PropertyChanged += EmailChanged;
                return e;

            } while (true);
        }

        private bool UpdateEmailInstance(Email e)
        {
            do
            {
                (var sender, e.LastSync) = _server.GetEmailSender(e.Id);
                if (e.LastSync == DateTimeOffset.MinValue) return false;
                e.Sender = sender!;
                var (recipient, newLast) = _server.GetEmailRecipient(e.Id);
                if (newLast == DateTimeOffset.MinValue) return false;
                if (e.LastSync != newLast) continue;
                e.Recipient = recipient!;
                (var subject, newLast) = _server.GetEmailSubject(e.Id);
                if (newLast == DateTimeOffset.MinValue) return false;
                if (e.LastSync != newLast) continue;
                e.Subject = subject!;
                (var content, newLast) = _server.GetEmailContent(e.Id);
                if (newLast == DateTimeOffset.MinValue) return false;
                if (e.LastSync != newLast) continue;
                e.Content = content!;
                (var timestamp, newLast) = _server.GetEmailTimestamp(e.Id);
                if (newLast == DateTimeOffset.MinValue) return false;
                if (e.LastSync != newLast) continue;
                e.Timestamp = timestamp!;
                (var status, newLast) = _server.GetEmailStatus(e.Id);
                if (newLast == DateTimeOffset.MinValue) return false;
                if (e.LastSync != newLast) continue;
                e.IsUnread = status!;
                return true;

            } while (true);
        }
        private void RemoveEmail(Email em)
        {
            var emails = FindEmailCollection(em);
            emails?.Remove(em);
            em.PropertyChanged -= EmailChanged;
        }

        private ObservableCollection<Email>? FindEmailCollection(Email email)
        {
            List<Folder> toSearch = Folders.Reverse().ToList();
            while (toSearch.Count > 0)
            {
                var folder = toSearch.Last();
                toSearch.RemoveAt(toSearch.Count - 1);
                if (folder.Emails.Contains(email))
                    return folder.Emails;
                toSearch.AddRange(folder.SubFolders.Reverse());
            }

            return null;
        }

        private IEnumerable<Folder> AllSubFolders
        {
            get
            {
                List<Folder> toEnumerate = new List<Folder>(Folders.Reverse());
                while (toEnumerate.Count > 0)
                {
                    int last = toEnumerate.Count - 1;
                    var folder = toEnumerate[last];
                    yield return folder;
                    toEnumerate.RemoveAt(last);
                    toEnumerate.AddRange(folder.SubFolders.Reverse());
                }
            }
        }


        private Folder? FindFolder(int id)
        {
            return AllSubFolders.FirstOrDefault(f => f.Id == id);
        }
        private void EmailsChanged(object? sender, NotifyCollectionChangedEventArgs e)
        {
            if (_updateLock) return;
            if (sender is not ObservableCollection<Email> emails) return;
            var updateOld = _updateLock;
            _updateLock = true;
            try
            {
                if (e.Action == NotifyCollectionChangedAction.Move) return;
                //Emails should really be exposed via ReadOnlyObservableCollection<T>
                //And deleting emails should be enabled via separate method
                if (e.NewItems is { Count : > 0 })
                    throw new ArgumentException("Emails can only be added by the server!", nameof(e.NewItems));
                if (e.OldItems != null)
                    foreach (Email removed in e.OldItems)
                    {
                        _server.DeleteEmail(removed.Id);
                        removed.PropertyChanged -= EmailChanged;
                    }
            }
            finally
            {
                _updateLock = updateOld;
            }
        }

        private void SubFoldersChanged(object? sender, NotifyCollectionChangedEventArgs e)
        {
            if (_updateLock) return;
            if (sender is not ObservableCollection<Folder> folders) return;
            var updateOld = _updateLock;
            _updateLock = true;
            try
            {
                //Folders should really be exposed via ReadOnlyObservableCollection<T>
                if (e.Action == NotifyCollectionChangedAction.Move) return;
                if (e.NewItems is { Count: > 0} || e.OldItems is { Count: > 0})
                    throw new ArgumentException("Folders can only be added or removed by the server!", nameof(e.NewItems));
            }
            finally
            {
                _updateLock = updateOld;
            }
        }

        private bool CheckFolderSync(Folder f, DateTimeOffset prevSync, DateTimeOffset currentSync)
        {
            if (prevSync == DateTimeOffset.MinValue || prevSync != f.LastSync && !UpdateFolderInstance(f))
            {
                RemoveFolder(f);
                return false;
            }
            f.LastSync = currentSync;
            return true;
        }

        private void FolderChanged(object? sender, PropertyChangedEventArgs e)
        {
            if (_updateLock) return;
            if (sender is not Folder f) return;
            var updateOld = _updateLock;
            _updateLock = true;
            try
            {
                switch (e.PropertyName)
                {
                    case nameof(Folder.Name):
                    {
                        var (prevSync, lastSync) = _server.SetFolderName(f.Id, f.Name);
                        CheckFolderSync(f, prevSync, lastSync);
                        break;
                    }
                }
            }
            finally
            {
                _updateLock = updateOld;
            }
        }

        private void EmailChanged(object? sender, PropertyChangedEventArgs e)
        {
            if (_updateLock) return;
            if (sender is not Email email) return;
            var updateOld = _updateLock;
            _updateLock = true;
            try
            {
                switch (e.PropertyName)
                {
                    case nameof(Email.Recipient):
                        PushEmailProperty(email, em => _server.SetEmailRecipient(em.Id, em.Recipient));
                        break;
                    case nameof(Email.Sender):
                        PushEmailProperty(email, em=>_server.SetEmailSender(em.Id, em.Sender));
                        break;
                    case nameof(Email.Subject):
                        PushEmailProperty(email, em=>_server.SetEmailSubject(em.Id, em.Subject));
                        break;
                    case nameof(Email.Content):
                        PushEmailProperty(email, em => _server.SetEmailContent(em.Id, em.Content));
                        break;
                    case nameof(Email.Timestamp):
                        throw new ArgumentException("Timestamp can only be changed by server!", nameof(e.PropertyName));
                    case nameof(Email.IsUnread):
                        PushEmailProperty(email, em => _server.SetEmailStatus(em.Id, em.IsUnread));
                        break;
                }
            }
            finally
            {
                _updateLock = updateOld;
            }

            void PushEmailProperty(Email em, Func<Email, Response<DateTimeOffset>> pushF)
            {
                var (prevSync, lastSync) = pushF(em);
                if (lastSync == DateTimeOffset.MinValue || prevSync != em.LastSync && !UpdateEmailInstance(em))
                    RemoveEmail(em);
                em.LastSync = lastSync;
            }
        }
    }
}
