﻿using EmailClient.Data;
using EmailClient.DomainModel;
using EmailClient.MVC.View;

namespace EmailClient.MVC.Controller
{
    public class AccountController : IAccountController
    {
        public AccountController(MailServer server)
        {
            Account = new Account(server);
        }

        public IAccountModel Account { get; }
        public void Refresh(Folder folder)
        {
            Account.RefreshFolder(folder);
        }

        public void RefreshAll()
        {
            Account.RefreshAll();
        }

        public void ComposeNewEmail(Form? owner)
        {
            var email = Account.CreateDraft();
            if (email != null)
                ShowEmailEditView(email, owner);
        }

        public void ComposeReplyEmail(Email email, Form? owner)
        {
            var reply = Account.CreateDraft();
            if (reply == null) return;
            reply.Subject = $"Re: {email.Subject}";
            reply.Recipient = email.Sender;
            reply.Content =
                $"{email.Recipient} at {email.Timestamp:yyyy-MM-dd HH:mm} sent:\n{email.Content}"
                    .ReplaceLineEndings(Environment.NewLine + "> ") + Environment.NewLine;
            ShowEmailEditView(reply, owner);
        }

        public void ComposeForwardEmail(Email email, Form? owner)
        {
            var subject = $"Re: {email.Subject}";
            var content = $"----\n{email.Recipient} at {email.Timestamp:yyyy-MM-dd HH:mm} sent:\n{email.Content}\n----\n"
                .ReplaceLineEndings();
            var reply = Account.CreateDraft(subject, content);
            if (reply == null) return;
                
            ShowEmailEditView(reply, owner);
        }

        public void ShowEmailDetails(Email email, Form? owner)
        {
            if (Account.DraftFolder?.Emails.Contains(email) == true)
                ShowEmailEditView(email, owner);
            else
                ShowEmailView(email, owner);

        }

        private class EmailController : IEmailController
        {
            private readonly IAccountController _parent;
            public Email Email { get; }

            public EmailController(IAccountController parent, Email email)
            {
                _parent = parent;
                Email = email;
            }

            public void ComposeReplyEmail(Form? owner)
            {
                _parent.ComposeReplyEmail(Email, owner);
            }

            public void ComposeForwardEmail(Form? owner)
            {
                _parent.ComposeReplyEmail(Email, owner);
            }
        }

        private void ShowEmailView(Email email, Form? owner)
        {
            var v = new EmailView(new EmailController(this, email));
            v.Owner = owner;
            v.Show();
        }

        private class EmailEditController : IEmailEditController
        {
            private readonly IAccountController _parent;
            private bool _isSent = false;

            public EmailEditController(IAccountController parent, Email email)
            {
                _parent = parent;
                Email = email;
            }

            public Email Email { get; }

            public void Send()
            {
                _parent.Account.SendEmail(Email, true);
                _isSent = true;
            }

            public bool CloseUnsent(Form? owner)
            {
                if (_isSent) return true;
                var dialog = new CloseConfirmation(owner);
                var result = dialog.Show();
                if (result == DialogResult.Cancel)
                    return false;
                if (result == DialogResult.No)
                    _parent.Account.DeleteDraft(Email);
                return true;
            }
        }

        private void ShowEmailEditView(Email email, Form? owner)
        {
            var v = new EmailEditView(new EmailEditController(this, email));
            v.Owner = owner;
            v.Show();
        }

        public void SetAsRead(Email email)
        {
            email.IsUnread = false;
        }
    }
}
