﻿using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;
using EmailClient.DomainModel;

namespace EmailClient.MVC.View
{
    public interface IAccountController
    {
        IAccountModel Account { get; }
        void Refresh(Folder folder);
        void RefreshAll();
        void ComposeNewEmail(Form? owner);
        void ComposeReplyEmail(Email email, Form? owner);
        void ComposeForwardEmail(Email email, Form? owner);
        void ShowEmailDetails(Email email, Form? owner);
        void SetAsRead(Email email);
    }

    public partial class AccountView : Form
    {
        private readonly IAccountController _controller;
        private readonly IAccountModel _model;

        private readonly Font _boldItemFont;

        private readonly string _timestampFormat = "yyyy-MM-dd HH:mm";

        public AccountView(IAccountController controller)
        {
            InitializeComponent();
            _boldItemFont = new Font(lvFolderEmails.Font, FontStyle.Bold);
            _controller = controller;
            _model = _controller.Account;
            AddFolderTreeNodes(tvAccountFolders.Nodes, _model.Folders);
            _controller.Account.Folders.CollectionChanged += UpdateSubfolders;
        }

        #region Folder Tree View Handling

        private void AddFolderTreeNodes(TreeNodeCollection nodes, IEnumerable<Folder> folders)
        {
            foreach (var folder in folders)
                nodes.Add(CreateFolderTreeNode(folder));
        }

        private TreeNode CreateFolderTreeNode(Folder folder)
        {
            var node = new TreeNode(folder.Name);
            AddFolderTreeNodes(node.Nodes, folder.SubFolders);
            folder.PropertyChanged += UpdateFolder;
            folder.SubFolders.CollectionChanged += UpdateSubfolders;
            node.Tag = folder;
            return node;
        }

        private void UpdateFolder(object? sender, PropertyChangedEventArgs e)
        {
            Folder? f = sender as Folder;
            if (f == null) return;
            TreeNode? ftn = FindTreeNode(tvAccountFolders.Nodes, f);
            if (ftn == null) return;
            ftn.Name = f.Name;
        }

        private void UpdateSubfolders(object? sender, NotifyCollectionChangedEventArgs e)
        {
            //We will the order of elements
            if (e.Action == NotifyCollectionChangedAction.Move)
                return;
            TreeNodeCollection? nodes = FindTreeNodesCollection((ObservableCollection<Folder>)sender!);
            if (nodes != null)
                return;
            if (e.NewItems != null)
                foreach (Folder f in e.NewItems)
                    nodes!.Add(CreateFolderTreeNode(f));
            if (e.OldItems != null)
            {
                for (int i = 0; i < nodes!.Count; i++)
                    if (e.OldItems.Contains(nodes[i].Tag))
                    {
                        nodes.RemoveAt(i);
                        --i;
                    }
                foreach (Folder f in e.OldItems)
                {
                    UnlinkHandlers(f);
                    if (f == lvFolderEmails.Tag)
                        ClearEmailList();
                }
            }
        }

        private void UnlinkHandlers(Folder folder)
        {
            folder.SubFolders.CollectionChanged -= UpdateSubfolders;
            folder.PropertyChanged -= UpdateFolder;
            foreach (var subFolder in folder.SubFolders)
                UnlinkHandlers(subFolder);
        }

        private TreeNode? FindTreeNode(TreeNodeCollection nodes, Folder folder)
        {
            foreach (TreeNode node in nodes)
            {
                if (node.Tag == folder) return node;
                var subNode = FindTreeNode(node.Nodes, folder);
                if (subNode != null) return subNode;
            }
            return null;
        }

        /// <summary>
        /// Finds the Nodes collection that contains tree nodes corresponding to the given set of folders.
        ///
        /// If this is the collection of top-level folders in the account model, it will return the Nodes of
        /// the tree view. If this is a set of subfolders, the function will return the Nodes collection of the
        /// TreeNode corresponding to the parent folder of the subfolders.
        /// </summary>
        /// <remarks>
        /// Comparison is made only based on the reference to the ObservableCollection. If it is the same as the Folders
        /// property on the account model, TreeView nodes are returned. Otherwise each TreeNode stores the reference
        /// to a corresponding folder and the function compares their SubFolder property with <paramref name="folders"/>
        /// </remarks>
        /// <param name="folders">Set of folders for which to find a corresponding Nodes.</param>
        /// <returns>TreeNodeCollection which contains nodes corresponding to given set of folders.</returns>
        private TreeNodeCollection? FindTreeNodesCollection(ObservableCollection<Folder> folders)
        {
            if (folders == _model.Folders)
                return tvAccountFolders.Nodes;
            return FindTreeNodesCollection(tvAccountFolders.Nodes, folders);
        }

        private TreeNodeCollection? FindTreeNodesCollection(TreeNodeCollection nodes, ObservableCollection<Folder> folders)
        {
            foreach (TreeNode node in nodes)
            {
                if (node.Tag is Folder folder && folder.SubFolders == folders)
                    return node.Nodes;
                var items = FindTreeNodesCollection(node.Nodes, folders);
                if (items != null)
                    return items;
            }
            return null;
        }

        #endregion

        #region Email List Handling

        private void SetEmailList(Folder? f)
        {
            ClearEmailList();
            btnRefreshFolder.Enabled = (f != null);
            if (f == null)
                return;
            lvFolderEmails.Tag = f;
            foreach (var e in f.Emails)
                lvFolderEmails.Items.Add(CreateEmailListItem(e));
            f.Emails.CollectionChanged += UpdateEmailList;
        }

        private void ClearEmailList()
        {
            if (lvFolderEmails.Tag is Folder f)
                f.Emails.CollectionChanged -= UpdateEmailList;
            lvFolderEmails.Tag = null;
            foreach (ListViewItem item in lvFolderEmails.Items)
                if (item.Tag is Email e)
                    e.PropertyChanged -= UpdateEmail;
            lvFolderEmails.Items.Clear();
            tbEmailContent.Clear();
        }

        private ListViewItem CreateEmailListItem(Email email)
        {
            var item = new ListViewItem(new[]
            {
                email.IsUnread ? "*" : string.Empty,
                email.Subject,
                email.Recipient,
                email.Timestamp.ToString(_timestampFormat)
            }) {
                Tag = email
            };
            if (email.IsUnread)
                item.Font = _boldItemFont;
            email.PropertyChanged += UpdateEmail;
            return item;
        }

        private void UpdateEmailList(object? sender, NotifyCollectionChangedEventArgs e)
        {
            //Ignore changes in item order
            if (e.Action == NotifyCollectionChangedAction.Move)
                return;
            if (e.OldItems != null)
                foreach (Email email in e.OldItems)
                {
                    email.PropertyChanged -= UpdateEmail;
                    var item = FindListItem(email);
                    if (item != null)
                        lvFolderEmails.Items.Remove(item);
                    if (tbEmailContent.Tag == email)
                        tbEmailContent.Clear();
                }
            if (e.NewItems != null)
                foreach (Email email in e.NewItems)
                    lvFolderEmails.Items.Add(CreateEmailListItem(email));
        }

        private ListViewItem? FindListItem(Email e)
        {
            foreach (ListViewItem item in lvFolderEmails.Items)
                if (item.Tag == e)
                    return item;
            return null;
        }

        #endregion

        #region Email Handling

        private void UpdateEmail(object? sender, PropertyChangedEventArgs e)
        {
            Email? email = sender as Email;
            if (email == null) return;
            ListViewItem? item = FindListItem(email);
            if (item == null) return;
            switch (e.PropertyName)
            {
                case nameof(Email.IsUnread):
                    item.SubItems[0].Text = email.IsUnread ? "*" : string.Empty;
                    item.Font = email.IsUnread ? _boldItemFont : lvFolderEmails.Font;
                    break;
                case nameof(Email.Subject):
                    item.SubItems[1].Text = email.Subject;
                    break;
                case nameof(email.Recipient):
                    item.SubItems[2].Text = email.Recipient;
                    break;
                case nameof(Email.Timestamp):
                    item.SubItems[3].Text = email.Timestamp.ToString(_timestampFormat);
                    break;
                case nameof(Email.Content):
                    if (tbEmailContent.Tag == email)
                        tbEmailContent.Text = email.Content;
                    break;
            }
        }

        private void SetEmail(Email? e)
        {
            tbEmailContent.Tag = e;
            tbEmailContent.Text = e?.Content;
            if (e != null)
                _controller.SetAsRead(e);
            btnReplyTo.Enabled = btnForwardEmail.Enabled = (e != null);

        }

        #endregion

        #region Event Handlers

        private void btnRefreshFolder_Click(object sender, EventArgs e)
        {
            if (tvAccountFolders.SelectedNode is { Tag: Folder f })
                _controller.Refresh(f);
        }

        private void btnNewEmail_Click(object sender, EventArgs e)
        {
            _controller.ComposeNewEmail(this);
        }

        private void btnReplyTo_Click(object sender, EventArgs e)
        {
            if (tbEmailContent.Tag is Email email)
                _controller.ComposeReplyEmail(email, this);
        }

        private void tvAccountFolders_AfterSelect(object sender, TreeViewEventArgs e)
        {
            SetEmailList(e.Node?.Tag as Folder);
        }

        private void btnForwardEmail_Click(object sender, EventArgs e)
        {
            if (tbEmailContent.Tag is Email email)
                _controller.ComposeForwardEmail(email, this);
        }

        private void lvFolderEmails_SelectedIndexChanged(object sender, EventArgs e)
        {
            Email? email = null;
            if (lvFolderEmails.SelectedItems.Count > 0)
                email = lvFolderEmails.SelectedItems[0].Tag as Email;
            SetEmail(email);
        }

        private void lvFolderEmails_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            ListViewItem? item = lvFolderEmails.HitTest(e.Location).Item;
            if (item is { Tag: Email email })
                _controller.ShowEmailDetails(email, this);
        }

        private void tRefreshAll_Tick(object sender, EventArgs e)
        {
            _controller.RefreshAll();
        }

        #endregion
    }
}
