﻿using EmailClient.MVP.View;
using EmailClient.Transactional.Entity;
using EmailClient.Transactional.Service;

namespace EmailClient.MVP.Presenter
{
    public class AccountPresenter
    {
        private readonly IAccountView _view;
        private readonly IAccountService _modelService;
        private readonly List<Folder> _model;
        private DateTimeOffset _modelLastSync;
        private int _sentFolderId = -1;
        private int _draftFolderId = -1;
        private readonly string _draftFolderName = "Drafts";
        private readonly string _sentFolderName = "Sent";

        public AccountPresenter(IAccountView view, IAccountService modelService)
        {
            _model = new List<Folder>();
            _modelLastSync = DateTimeOffset.MinValue;
            _modelService = modelService;
            _modelService.FolderAdded += AddFolderToView;
            _modelService.FolderRemoved += RemoveFolderFromView;
            _modelService.FolderChanged += UpdateFolderInView;
            _modelService.EmailAdded += AddEmailToView;
            _modelService.EmailRemoved += RemoveEmailFromView;
            _modelService.EmailChanged += UpdateEmailInView;
            _view = view;
            _view.ComposeEmailExecuted += CreateEmptyDraft;
            _view.ReplyToExecuted += ReplyToSelected;
            _view.ForwardToExecuted += ForwardSelected;
            _view.SelectedFolderChanged += ChangeCurrentFolderInView;
            _view.SelectedEmailChanged += ChangeCurrentEmailInView;
            _view.RefreshFolderExecuted += RefreshCurrentFolder;
            _view.RefreshAccountExecuted += RefreshModel;
            _view.ShowEmailExecuted += ShowEmailDetails;
            RefreshModelImpl();
        }

        private void RefreshModelImpl()
        {
            _modelLastSync = _modelService.RefreshAll(_model, _modelLastSync);
        }

        private void ShowEmailDetails(object? sender, EventArgs e)
        {
            var email = SelectedEmail;
            if (!email.HasValue) return;
            if (_view.SelectedFolderId == _draftFolderId)
                ShowDraftEdit(email.Value);
            else
                ShowEmail(email.Value, _view.SelectedFolderId);
        }

        private void RefreshModel(object? sender, EventArgs e)
        {
            RefreshModelImpl();
        }

        private void RefreshCurrentFolder(object? sender, EventArgs e)
        {
            var folderId = _view.SelectedFolderId;
            RefreshFolderImpl(folderId);
        }

        private void RefreshFolderImpl(int folderId)
        {
            var (folder, parentId) = FindFolder(folderId);
            if (folder.HasValue)
                _modelService.RefreshFolder(folder.Value, parentId);
        }

        private void ChangeCurrentEmailInView(object? sender, EventArgs e)
        {
            Folder? f = FindFolder(_view.SelectedFolderId).Folder;
            Email? email = null;
            int emailId = _view.SelectedEmailId;
            if (f != null)
                for (var i = 0; i < f.Value.Emails.Count; i++)
                    if (f.Value.Emails[i].Id == emailId)
                    {
                        if (f.Value.Emails[i].IsUnread)
                        {
                            f.Value.Emails[i] =
                                _modelService.ChangeEmailReadStatus(f.Value.Emails[i] with { IsUnread = false },
                                    f.Value.ID);
                        }

                        email = f.Value.Emails[i];
                    }

            bool emailSelected = email.HasValue;
            _view.EmailContent = emailSelected ? email!.Value.Content : string.Empty;
            _view.DraftCommandsEnabled = emailSelected;
        }

        private (Folder? Folder, int ParentId) FindFolder(int folderId)
        {
            return FindFolder(folderId, _model, -1);
        }

        private (Folder? Folder, int ParentId) FindFolder(int folderId, List<Folder> subFolders, int parentId)
        {
            foreach (var folder in subFolders)
            {
                if (folder.ID == folderId)
                    return (folder, parentId);
                var subFolder = FindFolder(folderId, folder.SubFolders, folder.ID);
                if (subFolder.Folder.HasValue) return subFolder;
            }
            return (null, -1);
        }

        private void ChangeCurrentFolderInView(object? sender, EventArgs e)
        {
            _view.ClearEmails();
            var f = FindFolder(_view.SelectedFolderId);
            _view.RefreshCommandEnabled = f.Folder.HasValue;
            if (!f.Folder.HasValue) return;
            f.Folder = _modelService.RefreshFolder(f.Folder.Value, f.ParentId);
            foreach (var email in f.Folder.Value.Emails)
                AddEmailToViewImpl(email);
        }

        private void ForwardSelected(object? sender, EventArgs e)
        {
            var toForward = SelectedEmail;

            if (!toForward.HasValue) return;
            CreateForwardDraftImpl(toForward.Value);
        }

        private void CreateForwardDraftImpl(Email toForward)
        {
            var subject = $"Re: {toForward.Subject}";
            var content =
                $"----\n{toForward.Recipient} at {toForward.Timestamp:yyyy-MM-dd HH:mm} sent:\n{toForward.Content}\n----\n"
                    .ReplaceLineEndings();
            CreateDraftImpl(subject, content, string.Empty);
        }

        private Email? SelectedEmail
        {
            get
            {
                var folderId = _view.SelectedFolderId;
                var emailId = _view.SelectedEmailId;
                var (folder, _) = FindFolder(folderId);
                if (!folder.HasValue) return null;
                Email? selected = null;
                foreach (var email in folder.Value.Emails)
                    if (email.Id == emailId)
                    {
                        selected = email;
                        break;
                    }

                return selected;
            }
        }

        private void CreateDraftImpl(string subject, string content, string recipient)
        {
            var draftEmail = _modelService.AddEmail(_draftFolderId,
                _modelService.GetAccountAddress(),
                recipient,
                subject,
                content);
            if (!draftEmail.HasValue) return;
            if (_view.SelectedFolderId == _draftFolderId)
                RefreshFolderImpl(_draftFolderId);
            ShowDraftEdit(draftEmail.Value);
        }

        private void ShowDraftEdit(Email draftEmail)
        {
            var view = new EmailEditView(_view as Form);
            EmailEditPresenter presenter = new EmailEditPresenter(draftEmail, _draftFolderId, view);
            presenter.DeleteExecuted += DeleteDraft;
            presenter.SentExecuted += SendDraft;
            presenter.UpdateExecuted += UpdateDraft;
            view.Show();
        }

        private void UpdateDraft(object sender, EmailEventArgs e)
        {
            _modelService.ChangeEmail(e.Email, e.ParentFolderId);
            RefreshFolderImpl(_draftFolderId);
        }

        private void ShowEmail(Email email, int parentId)
        {
            var view = new EmailView(_view as Form);
            EmailPresenter presenter = new EmailPresenter(_modelService, email, parentId, view);
            presenter.ForwardEmailExecuted += CreateForwardDraft;
            presenter.ReplyToExecuted += CreateReplyDraft;
            view.Show();
        }

        private void CreateForwardDraft(object sender, EmailEventArgs e)
        {
            CreateForwardDraftImpl(e.Email);
        }

        private void CreateReplyDraft(object sender, EmailEventArgs e)
        {
            CreateReplyDraftImpl(e.Email);
        }

        private void SendDraft(object sender, EmailEventArgs e)
        {
            _modelService.SendEmail(e.Email);
            RefreshFolderImpl(_sentFolderId);
        }

        private void DeleteDraft(object sender, EmailEventArgs e)
        {
            _modelService.DeleteEmail(e.Email, e.ParentFolderId);
            RefreshFolderImpl(_draftFolderId);
        }

        private void ReplyToSelected(object? sender, EventArgs e)
        {
            var toReply = SelectedEmail;
            if (!toReply.HasValue) return;
            CreateReplyDraftImpl(toReply.Value);
        }

        private void CreateReplyDraftImpl(Email email)
        {
            var subject = $"Re: {email.Subject}";
            var content =
                $"{email.Recipient} at {email.Timestamp:yyyy-MM-dd HH:mm} sent:\n{email.Content}"
                    .ReplaceLineEndings(Environment.NewLine + "> ") + Environment.NewLine;
            CreateDraftImpl(subject, content,
                email.Sender == _modelService.GetAccountAddress()
                    ? email.Recipient
                    : email.Sender);
        }

        private void CreateEmptyDraft(object? sender, EventArgs e)
        {
            CreateDraftImpl(string.Empty, string.Empty, string.Empty);
        }

        private void UpdateEmailInView(object sender, EmailEventArgs e)
        {
            if (e.ParentFolderId != _view.SelectedFolderId)
                return;
            _view.UpdateEmail(e.Email.Id,
                e.Email.Subject,
                e.Email.Recipient == _modelService.GetAccountAddress()
                    ? e.Email.Sender
                    : e.Email.Recipient,
                e.Email.Timestamp.ToString("yyyy-MM-dd HH:ss"),
                e.Email.IsUnread);
        }

        private void RemoveEmailFromView(object sender, EmailEventArgs e)
        {
            if (_view.SelectedFolderId == e.ParentFolderId)
                _view.RemoveEmail(e.Email.Id);
        }

        private void AddEmailToView(object sender, EmailEventArgs e)
        {
            if (_view.SelectedFolderId == e.ParentFolderId)
                AddEmailToViewImpl(e.Email);
        }

        private void AddEmailToViewImpl(Email email)
        {
            _view.AddEmail(email.Id,
                email.Subject,
                email.Recipient == _modelService.GetAccountAddress()
                    ? email.Sender
                    : email.Recipient,
                email.Timestamp.ToString("yyyy-MM-dd HH:mm"),
                email.IsUnread);
        }

        private void UpdateFolderInView(object sender, FolderEventArgs e)
        {
            _view.UpdateFolder(e.Folder.ID, e.Folder.Name);
            if (e.Folder.Name == _draftFolderName)
                _draftFolderId = e.Folder.ID;
            else if (e.Folder.ID == _draftFolderId)
                _draftFolderId = -1;
            if (e.Folder.Name == _sentFolderName)
                _sentFolderId = e.Folder.ID;
            else if (e.Folder.ID == _sentFolderId)
                _sentFolderId = -1;
        }

        private void RemoveFolderFromView(object sender, FolderEventArgs e)
        {
            _view.RemoveFolder(e.Folder.ID, e.ParentFolderId);
            if (e.Folder.ID == _sentFolderId)
                _sentFolderId = -1;
            if (e.Folder.ID == _draftFolderId)
                _draftFolderId = -1;
        }

        private void AddFolderToView(object sender, FolderEventArgs e)
        {
            _view.AddFolder(e.Folder.ID, e.Folder.Name, e.ParentFolderId);
            if (e.Folder.Name == _draftFolderName)
                _draftFolderId = e.Folder.ID;
            else if (e.Folder.Name == _sentFolderName)
                _sentFolderId = e.Folder.ID;

        }
    }
}
