﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace EmailClient.MVP.View
{
    public interface IAccountView
    {
        event EventHandler? SelectedFolderChanged;
        event EventHandler? SelectedEmailChanged;
        event EventHandler? ComposeEmailExecuted;
        event EventHandler? ReplyToExecuted;
        event EventHandler? ForwardToExecuted;
        event EventHandler? RefreshFolderExecuted;
        event EventHandler? RefreshAccountExecuted;
        event EventHandler? ShowEmailExecuted;
        int SelectedFolderId { get; }
        int SelectedEmailId { get; }
        string EmailContent { get; set; }
        bool DraftCommandsEnabled { get; set; }
        bool RefreshCommandEnabled { get; set; }
        void AddFolder(int id, string name, int parentId = -1);
        void RemoveFolder(int id, int parentId = -1);
        void UpdateFolder(int id, string name);
        void ClearEmails();
        void AddEmail(int id, string subject, string address, string date, bool isUnread);
        void RemoveEmail(int id);
        void UpdateEmail(int id, string subject, string address, string date, bool isUnread);
    }

    public partial class AccountView : Form, IAccountView
    {
        private Font _boldItemFont;
        public AccountView()
        {
            InitializeComponent();
            _boldItemFont = new Font(lvFolderEmails.Font, FontStyle.Bold);
        }

        public event EventHandler? SelectedFolderChanged;
        public event EventHandler? SelectedEmailChanged;
        public event EventHandler? ComposeEmailExecuted;
        public event EventHandler? ReplyToExecuted;
        public event EventHandler? ForwardToExecuted;
        public event EventHandler? RefreshFolderExecuted;
        public event EventHandler? RefreshAccountExecuted;
        public event EventHandler? ShowEmailExecuted;

        private TreeNodeCollection? FindParentNodeCollection(int parentId)
        {
            if (parentId == -1)
                return tvAccountFolders.Nodes;
            var pNodes = tvAccountFolders.Nodes.Find(parentId.ToString(), true);
            return pNodes.Length > 0 ? pNodes[0].Nodes : null;
        }

        public void AddFolder(int id, string name, int parentId = -1)
        {
            var nodes = FindParentNodeCollection(parentId);
            if (nodes != null)
            {
                var node = new TreeNode(name)
                {
                    Name = id.ToString(),
                    Tag = id
                };
                nodes.Add(node);
            }
        }

        public void RemoveFolder(int id, int parentId = -1)
        {
            var nodes = FindParentNodeCollection(parentId);
            nodes?.RemoveByKey(id.ToString());
        }

        public void UpdateFolder(int id, string name)
        {
            var pNodes = tvAccountFolders.Nodes.Find(id.ToString(), true);
            if (pNodes.Length > 0)
                pNodes[0].Text = name;
        }

        public void ClearEmails()
        {
            lvFolderEmails.Items.Clear();
            tbEmailContent.Clear();
        }

        public void AddEmail(int id, string subject, string address, string date, bool isUnread)
        {
            var item = new ListViewItem(new[] { isUnread ? "*" : "", subject, address, date })
            {
                Name = id.ToString(),
                Tag = id,
                Font = ChooseItemFont(isUnread),
            };
            lvFolderEmails.Items.Add(item);
            lvFolderEmails.View = System.Windows.Forms.View.Details;
        }

        private Font ChooseItemFont(bool isUnread)
        {
            return isUnread ? _boldItemFont : lvFolderEmails.Font;
        }

        public void RemoveEmail(int id)
        {
            lvFolderEmails.Items.RemoveByKey(id.ToString());
        }

        public void UpdateEmail(int id, string subject, string address, string date, bool isUnread)
        {
            var items = lvFolderEmails.Items.Find(id.ToString(), false);
            if (items.Length > 0)
            {
                var item = items[0];
                item.SubItems[0].Text = isUnread ? "*" : "";
                item.SubItems[1].Text = subject;
                item.SubItems[2].Text = address;
                item.SubItems[3].Text = date;
                item.Font = ChooseItemFont(isUnread);
            }
        }

        public int SelectedFolderId => tvAccountFolders.SelectedNode is {Tag: int id and >= 0} ? id : -1;

        public int SelectedEmailId =>
            lvFolderEmails.SelectedItems.Count != 0 &&
            lvFolderEmails.SelectedItems[0].Tag is int id and >= 0
                ? id
                : -1;

        public string EmailContent
        {
            get => tbEmailContent.Text;
            set => tbEmailContent.Text = value;
        }

        public bool DraftCommandsEnabled
        {
            get => btnReplyTo.Enabled;
            set => btnReplyTo.Enabled = btnForwardEmail.Enabled = value;
        }

        public bool RefreshCommandEnabled
        {
            get => btnRefreshFolder.Enabled;
            set => btnRefreshFolder.Enabled = value;
        }

        protected virtual void OnSelectedFolderChanged()
        {
            SelectedFolderChanged?.Invoke(this, EventArgs.Empty);
        }

        protected virtual void OnSelectedEmailChanged()
        {
            SelectedEmailChanged?.Invoke(this, EventArgs.Empty);
        }
        protected virtual void OnComposeEmailExecuted()
        {
            ComposeEmailExecuted?.Invoke(this, EventArgs.Empty);
        }

        protected virtual void OnReplyToExecuted()
        {
            ReplyToExecuted?.Invoke(this, EventArgs.Empty);
        }

        protected virtual void OnForwardToExecuted()
        {
            ForwardToExecuted?.Invoke(this, EventArgs.Empty);
        }

        protected virtual void OnRefreshFolderExecuted()
        {
            RefreshFolderExecuted?.Invoke(this, EventArgs.Empty);
        }

        protected virtual void OnRefreshAccountExecuted()
        {
            RefreshAccountExecuted?.Invoke(this, EventArgs.Empty);
        }

        protected virtual void OnShowEmailExecuted()
        {
            ShowEmailExecuted?.Invoke(this, EventArgs.Empty);
        }

        private void btnRefreshFolder_Click(object sender, EventArgs e)
        {
            OnRefreshFolderExecuted();
        }

        private void btnNewEmail_Click(object sender, EventArgs e)
        {
            OnComposeEmailExecuted();
        }

        private void btnReplyTo_Click(object sender, EventArgs e)
        {
            OnReplyToExecuted();
        }

        private void btnForwardEmail_Click(object sender, EventArgs e)
        {
            OnForwardToExecuted();
        }

        private void tvAccountFolders_AfterSelect(object sender, TreeViewEventArgs e)
        {
            OnSelectedFolderChanged();
        }

        private void lvFolderEmails_SelectedIndexChanged(object sender, EventArgs e)
        {
            OnSelectedEmailChanged();
        }

        private void tRefreshAll_Tick(object sender, EventArgs e)
        {
            OnRefreshAccountExecuted();
        }

        private void lvFolderEmails_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            var info = lvFolderEmails.HitTest(e.Location);
            if (info.Item is {Tag: >= 0})
                OnShowEmailExecuted();
        }
    }
}
