﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Configuration;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using EmailClient.MVVM.ViewModel;

namespace EmailClient.MVVM.Behaviors
{
    public static class SaveDialog
    {
        public static DependencyProperty SaveCommandProperty = DependencyProperty.RegisterAttached("SaveCommand",
            typeof(ICommand), typeof(SaveDialog), new PropertyMetadata(null, EnsureHandlerSet));

        public static DependencyProperty DiscardCommandProperty = DependencyProperty.RegisterAttached("DiscardCommand",
            typeof(ICommand), typeof(SaveDialog), new PropertyMetadata(null, EnsureHandlerSet));

        public static DependencyProperty CaptionProperty = DependencyProperty.RegisterAttached("Caption",
            typeof(string), typeof(SaveDialog), new PropertyMetadata(string.Empty));

        public static DependencyProperty TextProperty = DependencyProperty.RegisterAttached("Text",
            typeof(string), typeof(SaveDialog), new PropertyMetadata(string.Empty));

        private static DependencyPropertyKey HasDialogPropertyKey =
            DependencyProperty.RegisterAttachedReadOnly("HasDialog", typeof(bool), typeof(SaveDialog),
                new PropertyMetadata(false, AddClosingHandler));

        public static DependencyProperty HasDialogProperty = HasDialogPropertyKey.DependencyProperty;

        private static void AddClosingHandler(DependencyObject dependencyObject, DependencyPropertyChangedEventArgs e)
        {
            if (dependencyObject is not Window w)
                throw new ArgumentException("Expected a Window", nameof(dependencyObject));
            if (e.OldValue == e.NewValue)
                return;
            if ((bool)e.NewValue)
                w.Closing += ConfirmWindowClosing;
            else
                w.Closing -= ConfirmWindowClosing;
        }

        private static bool HasProperty(this DependencyObject dependencyObject, DependencyProperty property)
        {
            var r = DependencyPropertyHelper.GetValueSource(dependencyObject, property);
            return r is
            {
                BaseValueSource: BaseValueSource.Default, IsAnimated: false, IsCoerced: false, IsCurrent: false
            };
        }

        private static bool HasAnyProperties(this DependencyObject dependencyObject)
        {
            return dependencyObject.HasProperty(SaveCommandProperty) ||
                   dependencyObject.HasProperty(DiscardCommandProperty);
        }

        public static bool GetHasDialog(DependencyObject dependencyObject)
        {
            return (bool)dependencyObject.GetValue(HasDialogProperty);
        }

        private static void SetHasDialog(DependencyObject dependencyObject, bool value)
        {
            if (value == GetHasDialog(dependencyObject))
                return;
            if (value)
                dependencyObject.SetValue(HasDialogPropertyKey, true);
            else
                dependencyObject.ClearValue(HasDialogPropertyKey);
        }

        public static ICommand? GetSaveCommand(DependencyObject dependencyObject)
        {
            return dependencyObject.GetValue(SaveCommandProperty) as ICommand;
        }

        public static void SetSaveCommand(DependencyObject dependencyObject, ICommand value)
        {
            dependencyObject.SetValue(SaveCommandProperty, value);
        }

        public static ICommand? GetDiscardCommand(DependencyObject dependencyObject)
        {
            return dependencyObject.GetValue(DiscardCommandProperty) as ICommand;
        }

        public static void SetDiscardCommand(DependencyObject dependencyObject, ICommand value)
        {
            dependencyObject.SetValue(DiscardCommandProperty, value);
        }

        public static string? GetCaption(DependencyObject dependencyObject)
        {
            return (string)dependencyObject.GetValue(CaptionProperty);
        }

        public static void SetCaption(DependencyObject dependencyObject, string? value)
        {
            dependencyObject.SetValue(CaptionProperty, value);
        }

        public static string? GetText(DependencyObject dependencyObject)
        {
            return (string)dependencyObject.GetValue(TextProperty);
        }

        public static void SetText(DependencyObject dependencyObject, string? value)
        {
            dependencyObject.SetValue(TextProperty, value);
        }

        private static void EnsureHandlerSet(DependencyObject dependencyObject, DependencyPropertyChangedEventArgs e)
        {
            SetHasDialog(dependencyObject, dependencyObject.HasAnyProperties());
        }

        private static void ConfirmWindowClosing(object? sender, CancelEventArgs e)
        {
            if (sender is not Window window || !window.HasAnyProperties()) return;

            var save = GetSaveCommand(window);
            var discard = GetDiscardCommand(window);
            if ((save == null || !save.CanExecute(null)) && (discard == null || !discard.CanExecute(null)))
                return;
            var text = GetText(window);
            if (string.IsNullOrEmpty(text))
                text = "Save changes?";
            var caption = GetCaption(window);
            if (string.IsNullOrEmpty(caption))
                caption = window.Title;
            var r = MessageBox.Show(window, text, caption, MessageBoxButton.YesNoCancel);
            switch (r)
            {
                case MessageBoxResult.Yes:
                    if (save is not null && save.CanExecute(null))
                        save.Execute(null);
                    break;
                case MessageBoxResult.No:
                    if (discard is not null && discard.CanExecute(null))
                        discard.Execute(null);
                    break;
                case MessageBoxResult.Cancel:
                    e.Cancel = true;
                    break;
            }
        }
    }
}
