﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using CommunityToolkit.Mvvm.ComponentModel;
using CommunityToolkit.Mvvm.Input;
using EmailClient.DomainModel;
using EmailClient.MVVM.Services;

namespace EmailClient.MVVM.ViewModel
{
    public partial class EmailViewModel : ObservableObject, IEquatable<EmailViewModel>
    {
        private Account _account;
        private Folder _folder;
        private Email _email;

        [ObservableProperty]
        private bool _isSelected;

        public EmailViewModel(Account account, Folder folder, Email email)
        {
            _account = account;
            _folder = folder;
            _email = email;
            _email.PropertyChanged += PropagatePropertyChanged;
        }

        private void PropagatePropertyChanged(object? sender, PropertyChangedEventArgs e)
        {
            OnPropertyChanged(e.PropertyName);
        }
        
        private void EditDraft(string subject, string content, string? recipient = null)
        {
            var email = _account.CreateDraft(subject, content, recipient);
            if (email != null)
                DialogServiceLocator.Instance.EditDetails(new EmailViewModel(_account, _account.DraftFolder!, email));
        }

        [RelayCommand]
        public void Reply()
        {
            var subject = $"Re: {_email.Subject}";
            var content =
                $"{_email.Recipient} at {_email.Timestamp:yyyy-MM-dd HH:mm} sent:\n{_email.Content}"
                    .ReplaceLineEndings(Environment.NewLine + "> ") + Environment.NewLine;
            EditDraft(subject, content, Correspondent);
        }

        [RelayCommand]
        public void Forward()
        {
            var subject = $"Re: {_email.Subject}";
            var content =
                $"----\n{_email.Recipient} at {_email.Timestamp:yyyy-MM-dd HH:mm} sent:\n{_email.Content}\n----\n"
                    .ReplaceLineEndings();
            EditDraft(subject, content);
        }

        [RelayCommand]
        public void ShowDetails()
        {
            if (_folder == _account.DraftFolder)
                DialogServiceLocator.Instance.EditDetails(this);
            else
                DialogServiceLocator.Instance.ShowDetails(this);
        }

        [RelayCommand(CanExecute = nameof(CanBeSent))]
        public void Send()
        {
            if (CanBeSent)
            {
                _account.SendEmail(_email, true);
                Recipient = string.Empty;
                OnPropertyChanged(nameof(UnsentDraft));
                OnEditClose();
            }
        }

        [RelayCommand]
        public void CancelEdit()
        {
            OnEditClose();
        }

        public bool UnsentDraft => IsDraft && _folder.Emails.Contains(_email);

        public event EventHandler? EditClose;

        public bool CanBeSent => IsDraft && !string.IsNullOrEmpty(Recipient);

        [RelayCommand(CanExecute = nameof(UnsentDraft))]
        public void DeleteDraft()
        {
            if (UnsentDraft)
                _account.DeleteDraft(_email);
        }

        public bool IsDraft => _folder == _account.DraftFolder;
        
        public string Recipient
        {
            get => _email.Recipient;
            set
            {
                _email.Recipient = value;
                OnPropertyChanged(nameof(Correspondent));
                OnPropertyChanged(nameof(CanBeSent));
                SendCommand.NotifyCanExecuteChanged();
            }
        }

        public string Sender
        {
            get => _email.Sender;
            set
            {
                _email.Sender = value;
                OnPropertyChanged(nameof(Correspondent));
            }
        }

        public string Correspondent => _email.Sender == _account.Address ? _email.Recipient : _email.Sender;

        public string Subject
        {
            get => _email.Subject;
            set => _email.Subject = value;
        }

        public string Content
        {
            get => _email.Content;
            set => _email.Content = value;
        }

        public DateTimeOffset Timestamp => _email.Timestamp;

        public bool IsUnread
        {
            get => _email.IsUnread;
            set
            {
                _email.IsUnread = value;
                OnPropertyChanged(nameof(UnreadIndicator));
                OnPropertyChanged(nameof(ItemWeight));
            }
        }

        partial void OnIsSelectedChanged(bool value)
        {
            if (value && IsUnread)
                IsUnread = false;
        }

        public string UnreadIndicator => _email.IsUnread ? "*" : "";

        public FontWeight ItemWeight => _email.IsUnread ? FontWeights.Bold : FontWeights.Normal;

        internal Email Model => _email;

        protected virtual void OnEditClose()
        {
            EditClose?.Invoke(this, EventArgs.Empty);
        }

        public bool Equals(EmailViewModel? other)
        {
            return _email.Equals(other?._email);
        }

        public override bool Equals(object? obj)
        {
            return obj is EmailViewModel model && Equals(model);

        }

        public override int GetHashCode()
        {
            return _email.GetHashCode();
        }
    }
}
