﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using CommunityToolkit.Mvvm.ComponentModel;
using EmailClient.DomainModel;

namespace EmailClient.MVVM.ViewModel
{
    public partial class FolderViewModel : ObservableObject
    {
        private readonly Account _account;
        private readonly Folder _folder;
        
        private readonly ObservableCollection<EmailViewModel> _emailViewModels;
        private readonly ObservableCollection<FolderViewModel> _subFolderViewModels;

        [ObservableProperty]
        private bool _isSelected;

        public FolderViewModel(Account account, Folder folder)
        {
            _account = account;
            _folder = folder;
            _folder.PropertyChanged += PropagatePropertyChanged;
            _emailViewModels = new ObservableCollection<EmailViewModel>(_folder.Emails.Select(e =>
                new EmailViewModel(_account, _folder, e)));
            Emails = new(_emailViewModels);
            _subFolderViewModels =
                new ObservableCollection<FolderViewModel>(_folder.SubFolders.Select(f =>
                    new FolderViewModel(_account, f)));
            SubFolders = new(_subFolderViewModels);
            //Add synchronization between collections
            folder.SubFolders.CollectionChanged += SubFoldersChanged;
            folder.Emails.CollectionChanged += EmailsChanged;
        }

        private void EmailsChanged(object? sender, NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Add:
                {
                    var idx = e.NewStartingIndex;
                    foreach (Email email in e.NewItems!)
                        _emailViewModels.Insert(idx++, new EmailViewModel(_account, _folder, email));
                    break;
                }
                case NotifyCollectionChangedAction.Remove:
                {
                    var start = e.OldStartingIndex;
                    for (var index = 0; index < e.OldItems!.Count; ++index)
                        _emailViewModels.RemoveAt(start + index);

                    break;
                }
                case NotifyCollectionChangedAction.Replace:
                {
                    var idx = e.NewStartingIndex;
                    foreach (Email email in e.NewItems!)
                        _emailViewModels[idx++] = new EmailViewModel(_account, _folder, email);

                    break;
                }
                case NotifyCollectionChangedAction.Move when e.NewStartingIndex > e.OldStartingIndex:
                {
                    for (var index = e.OldItems!.Count - 1; index >= 0; --index)
                        _emailViewModels.Move(e.OldStartingIndex + index, e.NewStartingIndex + index);
                    break;
                }
                case NotifyCollectionChangedAction.Move:
                {
                    for (var index = 0; index < e.OldItems!.Count; ++index)
                        _emailViewModels.Move(e.OldStartingIndex + index, e.NewStartingIndex + index);
                    break;
                }
                case NotifyCollectionChangedAction.Reset:
                {
                    _subFolderViewModels.Clear();
                    break;
                }
            }
        }

        private void SubFoldersChanged(object? sender, NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Add:
                {
                    var idx = e.NewStartingIndex;
                    foreach (Folder subfolder in e.NewItems!)
                        _subFolderViewModels.Insert(idx++, new FolderViewModel(_account, subfolder));
                    break;
                }
                case NotifyCollectionChangedAction.Remove:
                {
                    var start = e.OldStartingIndex;
                    for (var index = 0; index < e.OldItems!.Count; ++index)
                        _subFolderViewModels.RemoveAt(start + index);

                    break;
                }
                case NotifyCollectionChangedAction.Replace:
                {
                    var idx = e.NewStartingIndex;
                    foreach (Folder subfolder in e.NewItems!)
                        _subFolderViewModels[idx++] = new FolderViewModel(_account, subfolder);

                    break;
                }
                case NotifyCollectionChangedAction.Move when e.NewStartingIndex > e.OldStartingIndex:
                {
                    for (var index = e.OldItems!.Count - 1; index >= 0; --index)
                        _subFolderViewModels.Move(e.OldStartingIndex + index, e.NewStartingIndex + index);
                    break;
                }
                case NotifyCollectionChangedAction.Move:
                {
                    for (var index = 0; index < e.OldItems!.Count; ++index)
                        _subFolderViewModels.Move(e.OldStartingIndex + index, e.NewStartingIndex + index);
                    break;
                }
                case NotifyCollectionChangedAction.Reset:
                {
                    _subFolderViewModels.Clear();
                    break;
                }
            }
        }

        private void PropagatePropertyChanged(object? sender, PropertyChangedEventArgs e)
        {
            OnPropertyChanged(e.PropertyName);
        }

        public string Name
        {
            get => _folder.Name;
            set => _folder.Name = value;
        }

        public ReadOnlyObservableCollection<EmailViewModel> Emails { get; }

        public ReadOnlyObservableCollection<FolderViewModel> SubFolders { get; }

        internal Folder Model => _folder;
    }
}
