﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using EmailClient.Data;
using EmailClient.Transactional.Entity;

namespace EmailClient.Transactional.Service
{
    public class FolderElementEventArgs : EventArgs
    {
        public FolderElementEventArgs(int parentFolderId)
        {
            ParentFolderId = parentFolderId;
        }

        public int ParentFolderId { get; set; }
    }

    public class FolderEventArgs : FolderElementEventArgs
    {
        public Folder Folder { get; }

        public FolderEventArgs(int parentFolderId, Folder folder) : base(parentFolderId)
        {
            Folder = folder;
        }

        public FolderEventArgs(Folder folder) : base(-1)
        {
            Folder = folder;
        }
    }

    public class EmailEventArgs : FolderElementEventArgs
    {
        public Email Email { get; }

        public EmailEventArgs(int parentFolderId, Email email) : base(parentFolderId)
        {
            Email = email;
        }
    }

    public delegate void FolderEventHandler(object sender, FolderEventArgs e);

    public delegate void EmailEventHandler(object sender, EmailEventArgs e);

    public interface IAccountService
    {
        event FolderEventHandler? FolderAdded;
        event FolderEventHandler? FolderRemoved;
        event FolderEventHandler? FolderChanged;
        event EmailEventHandler? EmailAdded;
        event EmailEventHandler? EmailRemoved;
        event EmailEventHandler? EmailChanged;
        string GetAccountAddress();

        Email? AddEmail(int parentId, string? sender = null, string? recipient = null, string? subject = null,
            string? content = null);

        bool DeleteEmail(Email e, int parentId);
        Email? SendEmail(Email e);
        DateTimeOffset RefreshAll(List<Folder> folders, DateTimeOffset lastSync);
        Folder RefreshFolder(Folder folder, int parentId);
        Folder RefreshFolderR(Folder folder, int parentId);
        Email ChangeEmail(Email email, int parentId);
        Email ChangeEmailReadStatus(Email email, int parentId);
    }

    public class Account : IAccountService
    {
        private readonly MailServer _server;

        public Account(MailServer server)
        {
            _server = server;
        }

        public event FolderEventHandler? FolderAdded;
        public event FolderEventHandler? FolderRemoved;
        public event FolderEventHandler? FolderChanged;
        public event EmailEventHandler? EmailAdded;
        public event EmailEventHandler? EmailRemoved;
        public event EmailEventHandler? EmailChanged;

        public string GetAccountAddress() => _server.Address;

        public Email? AddEmail(int parentId, string? sender = null, string? recipient = null, string? subject = null,
            string? content = null)
        {
            var ((emailId, folderLast), emailLast) = _server.AddEmail(parentId);
            if (folderLast == DateTimeOffset.MinValue)
                return null;
            if (!string.IsNullOrEmpty(sender))
                _server.SetEmailSender(emailId, sender);
            if (!string.IsNullOrEmpty(recipient))
                _server.SetEmailRecipient(emailId, recipient);
            if (!string.IsNullOrEmpty(subject))
                _server.SetEmailSubject(emailId, subject);
            if(!string.IsNullOrEmpty(content))
                _server.SetEmailContent(emailId, content);
            return CreateEmailInstance(emailId);
        }

        private void ChangeEmailImpl(Email email)
        {
            var (_, emailNew) = _server.SetEmailSender(email.Id, email.Sender);
            if (emailNew == DateTimeOffset.MinValue) return;
            (_, emailNew) = _server.SetEmailRecipient(email.Id, email.Recipient);
            if (emailNew == DateTimeOffset.MinValue) return;
            (_, emailNew) = _server.SetEmailSubject(email.Id, email.Subject);
            if (emailNew == DateTimeOffset.MinValue) return;
            (_, emailNew) = _server.SetEmailContent(email.Id, email.Content);
            if (emailNew == DateTimeOffset.MinValue) return;
            _server.SetEmailStatus(email.Id, email.IsUnread);
        }

        public Email ChangeEmail(Email email, int parentId)
        {
            ChangeEmailImpl(email);
            return RefreshEmail(email, parentId);
        }

        public Email ChangeEmailReadStatus(Email email, int parentId)
        {
            _server.SetEmailStatus(email.Id, email.IsUnread);
            return RefreshEmail(email, parentId);
        }

        public bool DeleteEmail(Email e, int parentId)
        {
            var (_, folderNew) = _server.DeleteEmail(e.Id);
            OnEmailRemoved(new (parentId, e));
            return folderNew != DateTimeOffset.MinValue;
        }

        public Email? SendEmail(Email e)
        {
            var ((emailId, _), emailLast) = _server.SendEmail(e.Subject, e.Content, e.Recipient);
            if (emailLast == DateTimeOffset.MinValue)
                return null;
            return CreateEmailInstance(emailId);
        }
        
        public DateTimeOffset RefreshAll(List<Folder> folders, DateTimeOffset lastSync)
        {
            var (fIds, newSync) = _server.GetFolders();
            if (newSync == DateTimeOffset.MinValue)
            {
                FoldersCleared(folders, -1);
                folders.Clear();
                return newSync;
            }
            
            RefreshSubfolders(folders, fIds!.ToList(), -1);

            return newSync;
        }


        public Folder RefreshFolder(Folder folder, int parentId)
        {
            var (name, newSync) = _server.GetFolderName(folder.ID);

            if (newSync == DateTimeOffset.MinValue)
            {
                FoldersCleared(folder.SubFolders, parentId);
                EmailsCleared(folder.Emails, folder.ID);
                folder.SubFolders.Clear();
                folder.Emails.Clear();
                return folder;
            }

            if (newSync != folder.LastSync)
            {
                folder.LastSync = newSync;
                if (newSync != DateTimeOffset.MinValue)
                    folder.Name = name!;
                OnFolderChanged(new(parentId, folder));
            }

            var (fIds, _) = _server.GetSubFolders(folder.ID);
            var folderIds = fIds!.ToList();

            RemoveSubfolders(folder.SubFolders, folderIds, folder.ID);
            AddSubfolders(folder.SubFolders, folderIds, folder.ID);

            RefreshEmails(folder.Emails, folder.ID);

            return folder;
        }

        public Folder RefreshFolderR(Folder folder, int parentId)
        {
            var (fIds, newSync) = _server.GetSubFolders(folder.ID);
            if (newSync == DateTimeOffset.MinValue)
            {
                FoldersCleared(folder.SubFolders, parentId);
                EmailsCleared(folder.Emails, folder.ID);
                folder.SubFolders.Clear();
                folder.Emails.Clear();
                return folder;
            }

            if (newSync != folder.LastSync)
            {
                folder.LastSync = newSync;
                (var name, newSync) = _server.GetFolderName(folder.ID);
                if (newSync != DateTimeOffset.MinValue)
                    folder.Name = name!;
                OnFolderChanged(new (parentId, folder));
            }

            RefreshSubfolders(folder.SubFolders, fIds!.ToList(), folder.ID);

            RefreshEmails(folder.Emails, folder.ID);

            return folder;
        }

        private void RefreshSubfolders(List<Folder> folders, List<int> folderIds, int parentId)
        {
            //remove folders
            RemoveSubfolders(folders, folderIds, parentId);

            //Check subfolders for updates
            for (var index = 0; index < folders.Count; index++)
                folders[index] = RefreshFolderR(folders[index], parentId);

            //add new folders
            AddSubfolders(folders, folderIds, parentId);
        }

        private void AddSubfolders(List<Folder> folders, List<int> folderIds, int parentId)
        {
            foreach (var folderId in folderIds.Where(fid => folders.All(sf => sf.ID != fid)))
                AddNewFolder(folders, parentId, folderId);
        }

        private void RemoveSubfolders(List<Folder> folders, List<int> folderIds, int parentId)
        {
            for (int i = 0; i < folders.Count; i++)
                if (!folderIds.Contains(folders[i].ID))
                    RemoveFolderAt(folders, i-- /*intentional post-decrement*/, parentId);
        }

        private void CreateSubFolders(List<Folder> folders, int parentId)
        {
            var (folderIds, newSync) = _server.GetSubFolders(parentId);
            if (newSync == DateTimeOffset.MinValue)
                return;
            foreach (var folderId in folderIds!)
                AddNewFolder(folders, parentId, folderId);
        }

        private void AddNewFolder(List<Folder> folders, int parentId, int folderId)
        {
            var (name, folderLast) = _server.GetFolderName(folderId);
            if (folderLast == DateTimeOffset.MinValue) return;
            var folder = new Folder(folderId, name!, folderLast);
            folders.Add(folder);
            OnFolderAdded(new(parentId, folder));
            CreateSubFolders(folder.SubFolders, folderId);
            CreateEmails(folder.Emails, folderId);
        }

        private void RemoveFolderAt(List<Folder> folders, int i, int parentId)
        {
            var folder = folders[i];
            FoldersCleared(folder.SubFolders, folder.ID);
            EmailsCleared(folder.Emails, folder.ID);
            folders.RemoveAt(i);
            OnFolderRemoved(new(parentId, folder));
        }

        private void FoldersCleared(List<Folder> folders, int parentId)
        {
            foreach (var folder in folders)
            {
                FoldersCleared(folder.SubFolders, folder.ID);
                EmailsCleared(folder.Emails, folder.ID);
                OnFolderRemoved(new (parentId, folder));
            }
            folders.Clear();
        }

        private void RefreshEmails(List<Email> emails, int parentId)
        {
            var (eIds, newSync) = _server.GetFolderEmails(parentId);
            if (newSync == DateTimeOffset.MinValue)
            {
                EmailsCleared(emails, parentId);
                emails.Clear();
                return;
            }

            var emailIds = eIds!.ToList();
            //remove emails
            for(int i = 0; i < emails.Count; i++)
                if (!emailIds.Contains(emails[i].Id))
                {
                    var email = emails[i];
                    emails.RemoveAt(i);
                    OnEmailRemoved(new (parentId, email));
                }
            //Check emails for updates
            for (int i = 0; i < emails.Count; i++)
                emails[i] = RefreshEmail(emails[i], parentId);
            //Add new emails
            foreach (var emailId in emailIds.Where(eid => emails.All(em => em.Id != eid)))
                AddNewEmail(emails, parentId, emailId);
        }

        private Email RefreshEmail(Email email, int parentId)
        {
            if (!_server.HasEmailChanged(email.Id, email.LastSync))
                return email;
            email = RefreshEmailImpl(email);
            OnEmailChanged(new (parentId, email));
            return email;
        }

        private Email RefreshEmailImpl(Email email)
        {
            var (recipient, emailNew) = _server.GetEmailRecipient(email.Id);
            if (emailNew == DateTimeOffset.MinValue) return email;
            email.Recipient = recipient!;
            (var sender, emailNew) = _server.GetEmailSender(email.Id);
            if (emailNew == DateTimeOffset.MinValue) return email;
            email.Sender = sender!;
            (var subject, emailNew) = _server.GetEmailSubject(email.Id);
            if (emailNew == DateTimeOffset.MinValue) return email;
            email.Subject = subject!;
            (var content, emailNew) = _server.GetEmailContent(email.Id);
            if (emailNew == DateTimeOffset.MinValue) return email;
            email.Content = content!;
            (var timestamp, emailNew) = _server.GetEmailTimestamp(email.Id);
            if (emailNew == DateTimeOffset.MinValue) return email;
            email.Timestamp = timestamp;
            (var status, emailNew) = _server.GetEmailStatus(email.Id);
            if (emailNew != DateTimeOffset.MinValue) email.IsUnread = status;

            return email;
        }

        private void CreateEmails(List<Email> emails, int parentId)
        {
            var (emailIds, newSync) = _server.GetFolderEmails(parentId);
            if (newSync == DateTimeOffset.MinValue)
                return;
            foreach (var emailId in emailIds!) AddNewEmail(emails, parentId, emailId);
        }

        private void AddNewEmail(List<Email> emails, int parentId, int emailId)
        {
            var email = CreateEmailInstance(emailId);
            if (!email.HasValue) return;
            emails.Add(email.Value);
            OnEmailAdded(new(parentId, email.Value));
        }

        private Email? CreateEmailInstance(int emailId)
        {
            var (recipient, emailLast) = _server.GetEmailRecipient(emailId);
            if (emailLast == DateTimeOffset.MinValue) return null;
            var (sender, emailNew) = _server.GetEmailSender(emailId);
            if (emailNew == DateTimeOffset.MinValue) return null;
            (var subject, emailNew) = _server.GetEmailSubject(emailId);
            if (emailNew == DateTimeOffset.MinValue) return null;
            (var content, emailNew) = _server.GetEmailContent(emailId);
            if (emailNew == DateTimeOffset.MinValue) return null;
            (var timestamp, emailNew) = _server.GetEmailTimestamp(emailId);
            if (emailNew == DateTimeOffset.MinValue) return null;
            (var status, emailNew) = _server.GetEmailStatus(emailId);
            if (emailNew == DateTimeOffset.MinValue) return null;
            return new Email(emailId, sender!, recipient!, subject!, content!, timestamp, status, emailLast);
        }

        private void EmailsCleared(List<Email> emails, int parentId)
        {
            foreach (var email in emails)
                OnEmailRemoved(new(parentId, email));
            emails.Clear();
        }

        protected virtual void OnFolderAdded(FolderEventArgs e)
        {
            FolderAdded?.Invoke(this, e);
        }

        protected virtual void OnFolderRemoved(FolderEventArgs e)
        {
            FolderRemoved?.Invoke(this, e);
        }

        protected virtual void OnFolderChanged(FolderEventArgs e)
        {
            FolderChanged?.Invoke(this, e);
        }

        protected virtual void OnEmailAdded(EmailEventArgs e)
        {
            EmailAdded?.Invoke(this, e);
        }

        protected virtual void OnEmailRemoved(EmailEventArgs e)
        {
            EmailRemoved?.Invoke(this, e);
        }

        protected virtual void OnEmailChanged(EmailEventArgs e)
        {
            EmailChanged?.Invoke(this, e);
        }
    }
}
